-----------------------------------------------------------------
-- Procedure BuildTree v1.0 PL/SQL                             --
-----------------------------------------------------------------
-- Builds a decision tree using the ID3 method (Quinlan, 1986) --
-----------------------------------------------------------------
-- (c) 2001 J. Darmont, ERIC, University of Lyon 2             --
-----------------------------------------------------------------
-- E-mail: jerome.darmont@univ-lyon2.fr                        --
-----------------------------------------------------------------

--
-- Input parameters:   table_name:	source table name
--                     class:		class attribute name
--                     res_name:	result table name
--                     min_gain:	(strict) minimum information gain for node building
--		       root_view:       root node view name
--                     del:		clean-up views after execution (True/False)
--

CREATE OR REPLACE PROCEDURE BuildTree(	table_name VARCHAR,
					class VARCHAR,
					res_name VARCHAR DEFAULT 'BTRES',
                                      	min_gain REAL DEFAULT 0,
					root_view VARCHAR DEFAULT 'BTROOT',
					del BOOLEAN DEFAULT TRUE) AUTHID CURRENT_USER AS

--		
-- Data types declaration
--
								-- Node
TYPE Node IS RECORD( num INTEGER,				-- Node number
		     nview VARCHAR(255),			-- Node view name
		     rule VARCHAR(30),				-- Node rule
		     entrop REAL,				-- Node entropy
		     pop INTEGER);				-- Node population strength

TYPE NodeList IS TABLE OF Node; 				-- List of nodes

TYPE StringList IS TABLE OF VARCHAR(255);			-- List of strings

TYPE DynCursor IS REF CURSOR;					-- Dynamic cursor

--
-- Variables declaration
--

stack NodeList := NodeList();		-- Stack of nodes (empty)
cnode Node;				-- Current node
att_names DynCursor;			-- List of attribute names
catt VARCHAR(30);			-- Current attribute name
att_values DynCursor;			-- List of current attribute values
cval VARCHAR(30);			-- Current value
egain REAL;				-- Information gain 
nnode Node;				-- New node
viewq VARCHAR(1000);			-- View creation query
att_names2 DynCursor;			-- List of attribute names
att_name VARCHAR(30);			-- Attribute name
candq VARCHAR(1000);			-- Candidate insertion query
maxg REAL;				-- Maximum information gain
max_var VARCHAR(30);			-- Variable associated to maximum information gain
candidates DynCursor;			-- Candidate list
cvar VARCHAR(30);			-- Attribute in candidate list
elist CNodeList;			-- List of nodes in candidate list
delist StringList := StringList();	-- List of nodes whose views must be deleted (empty)
resq VARCHAR(1000);			-- Result query
classval StringList := StringList();	-- Class attribute values list
nodenum INTEGER := 0;			-- Number of nodes in tree

i BINARY_INTEGER;
j BINARY_INTEGER;

--
-- Entropy and population computation procedure
--

PROCEDURE Entropy(view_name IN VARCHAR, -- Name of view associated to the treated node
                  ent OUT REAL, 	-- Entropy value
                  totpop OUT INTEGER)   -- Population total
IS

TYPE PopStrList IS TABLE OF INTEGER;	-- Type table of population strengths
popstr PopStrList := PopStrList();	-- Population strengths table
cstr INTEGER;				-- Current population strength
curs DynCursor;				-- Cursor for counting population
i BINARY_INTEGER;			-- Loop index

BEGIN

  -- Init.
  ent := 0;
  totpop := 0;
  
  -- Count population strengths and store them in popstr
  -- Total population strength is computed on the fly
  OPEN curs FOR 
    'SELECT COUNT(*) FROM ' || view_name || ' GROUP BY ' || class;
  LOOP
    FETCH curs INTO cstr;
    EXIT WHEN curs%NOTFOUND;
    popstr.EXTEND;
    popstr(popstr.COUNT) := cstr;
    totpop := totpop + cstr;
/*  Debug    
    DBMS_OUTPUT.PUT_LINE('popstr('||TO_CHAR(popstr.COUNT)||')='||TO_CHAR(popstr(popstr.COUNT)));
    DBMS_OUTPUT.PUT_LINE('totpop='||TO_CHAR(totpop));
*/
  END LOOP;
  CLOSE curs;
  
  -- Compute entropy
  FOR i IN 1..popstr.COUNT LOOP
    ent := ent - (popstr(i)/totpop)*LOG(2,(popstr(i)/totpop));
  END LOOP;
  ent := ROUND(ent,8); -- Precision adaptation
/*  Debug  
  DBMS_OUTPUT.PUT_LINE('Entropy = '||TO_CHAR(ent));
  DBMS_OUTPUT.PUT_LINE('Population = '||TO_CHAR(totpop));
*/
  
  popstr.TRIM(popstr.COUNT); -- Clean table

END;

--
-- Result output procedure
--

PROCEDURE Result( num INTEGER,		-- Node number
		  parent VARCHAR,	-- Parent node number or NULL
		  rule VARCHAR,		-- Building rule
		  view_name VARCHAR)     -- Node associated view
IS		 

curs DynCursor;				-- Cursor for counting population
resq VARCHAR(255);			-- Result query
v VARCHAR(30);				-- Class attribute value
p INTEGER;				-- Associated population
i BINARY_INTEGER;
j BINARY_INTEGER;

BEGIN

  -- Query init.	

  resq := 'INSERT INTO '||res_name||' VALUES('||TO_CHAR(num)||', '||parent||', ';
  IF rule <> 'NULL' THEN
    resq := resq || '''' || rule || '''';
  ELSE 
    resq := resq || 'NULL';
  END IF;
  i:=0;
  
  -- Compute populations and update query
  
  OPEN curs FOR 
    'SELECT '||class||', COUNT('||class||') FROM ' || view_name || ' GROUP BY ' || class || ' ORDER BY ' || class;
  LOOP
    FETCH curs INTO v, p;
    EXIT WHEN curs%NOTFOUND;
    i := i + 1;
    -- Debug DBMS_OUTPUT.PUT_LINE('v='||v||' p='||TO_CHAR(p)||' classval('||TO_CHAR(i)||')='||classval(i));
    WHILE (v <> classval(i)) LOOP
      resq := resq || ', 0';
      i := i + 1;
    END LOOP;
    resq := resq || ', ' || TO_CHAR(p);
  END LOOP;
  CLOSE curs;  
  
  -- Complete and execute query
  
  FOR j IN i+1..classval.COUNT LOOP
    resq := resq || ', 0';  	
  END LOOP;
  resq := resq || ')';
  -- Debug DBMS_OUTPUT.PUT_LINE(resq);
  EXECUTE IMMEDIATE resq;
  
END;

--	
-- Main program code
--

BEGIN

  -- Echo parameters
  
  DBMS_OUTPUT.PUT_LINE('table_name=' || table_name);
  DBMS_OUTPUT.PUT_LINE('class=' || class);
  DBMS_OUTPUT.PUT_LINE('res_name=' || res_name);
  DBMS_OUTPUT.PUT_LINE('min_gain=' || TO_CHAR(min_gain));
  DBMS_OUTPUT.PUT_LINE('root_view=' || root_view);
  IF del THEN
    DBMS_OUTPUT.PUT_LINE('del=TRUE');
  ELSE
    DBMS_OUTPUT.PUT_LINE('del=FALSE');
  END IF;  
   
  -- Create result table + list of class attribute values
  
  resq := 'CREATE TABLE '||res_name||'(node NUMBER(2) PRIMARY KEY, parent NUMBER(2), rule VARCHAR(30)';
  -- Fetch class attribute values
  OPEN att_values FOR 
    'SELECT DISTINCT '||class||' FROM '||table_name||' ORDER BY '||class; 
  LOOP      
    FETCH att_values INTO cval;
    EXIT WHEN att_values%NOTFOUND; -- Exit loop when last row is fetched			                   
    classval.EXTEND;
    classval(classval.COUNT) := cval;
    resq := resq || ', '||class||'_' || cval || ' NUMBER(5)';
  END LOOP; -- On att_values
  CLOSE att_values; 
  resq := resq || ', CONSTRAINT ' || res_name || '_H FOREIGN KEY(parent) REFERENCES ' || res_name || '(node))';    
  -- Debug DBMS_OUTPUT.PUT_LINE(resq);
  EXECUTE IMMEDIATE resq;
  DBMS_OUTPUT.PUT_LINE('Creating result table ' || res_name);
  
  -- Create starting node and stack it
  
  EXECUTE IMMEDIATE
    'CREATE VIEW ' || root_view || ' AS SELECT * FROM ' || table_name;
  -- Push
  stack.EXTEND;
  stack(stack.COUNT).num := nodenum;
  stack(stack.COUNT).nview := root_view;
  Entropy(root_view, stack(stack.COUNT).entrop, stack(stack.COUNT).pop);
  DBMS_OUTPUT.PUT_LINE('Creating node view ' || root_view);
  Result(nodenum, 'NULL', 'NULL', root_view); -- Output result
  nodenum := nodenum + 1; -- Next node
    
  -- Apply treatment while the stack is not empty
  
  WHILE stack.COUNT > 0 LOOP

    -- Pop current node
    
    cnode := stack(stack.COUNT);
    stack.TRIM;
    
/*  Debug    
    DBMS_OUTPUT.PUT_LINE('cnode.nview=' || cnode.nview);
    DBMS_OUTPUT.PUT_LINE('cnode.entrop=' || TO_CHAR(cnode.entrop));
    DBMS_OUTPUT.PUT_LINE('cnode.pop=' || TO_CHAR(cnode.pop));
*/
    
    -- Clean candidate list
    
    EXECUTE IMMEDIATE 'DELETE FROM BTCandidate';
    
    -- For each attribute but the class attribute
    
    OPEN att_names FOR 
      'SELECT COLUMN_NAME FROM USER_TAB_COLUMNS WHERE TABLE_NAME=UPPER('''||cnode.nview||''') AND COLUMN_NAME<>UPPER('''||class||''') ORDER BY COLUMN_NAME';
    
    LOOP
        
      -- Read current attribute name
         
      FETCH att_names INTO catt;
      EXIT WHEN att_names%NOTFOUND; -- Exit loop when last row is fetched 	           
      
      -- Debug DBMS_OUTPUT.PUT_LINE('catt=' || catt); 
      
      -- Create a candidate in list for current attribute
      
      candq := '';
      egain := cnode.entrop; -- Information gain is reset
      
      -- For each distinct value of current attribute
      
      OPEN att_values FOR 
        'SELECT DISTINCT '||catt||' FROM '||cnode.nview||' ORDER BY '||catt; 
        
      i := 0;        
      LOOP      
      
        -- Read current attribute value
          
        FETCH att_values INTO cval;
        EXIT WHEN att_values%NOTFOUND; -- Exit loop when last row is fetched	
        
        -- Debug DBMS_OUTPUT.PUT_LINE('cval=' || cval);
        
        -- Build new node
        
        i := i + 1;
        nnode.nview := cnode.nview || '_' || catt || TO_CHAR(i);
        viewq := 'CREATE VIEW ' || nnode.nview || ' AS SELECT ';
        nnode.rule := catt || '=' || cval;
        
        -- Select all attributes besides current attribute
        
        OPEN att_names2 FOR 
          'SELECT COLUMN_NAME FROM USER_TAB_COLUMNS WHERE TABLE_NAME=UPPER('''||cnode.nview||''') AND COLUMN_NAME<>UPPER('''||catt||''')';                
        j := 0;
        LOOP
          FETCH att_names2 INTO att_name;
          EXIT WHEN att_names2%NOTFOUND; -- Exit loop when last row is fetched	
          j := j + 1;
          IF j > 1 THEN
            viewq := viewq || ',';
          END IF;
          viewq := viewq || att_name;            
        END LOOP;
        CLOSE att_names2; 
        
        -- Conclude query and execute it
          
        viewq := viewq || ' FROM ' || cnode.nview || ' WHERE ' || catt || '=''' || cval || '''';
        -- Debug DBMS_OUTPUT.PUT_LINE(viewq);
        EXECUTE IMMEDIATE viewq;
        
        -- Compute entropy and update information gain
        
	Entropy(nnode.nview, nnode.entrop, nnode.pop);
        egain := egain - (nnode.pop/cnode.pop)*nnode.entrop;        
        
        -- Insert node in list
        
        IF i > 1 THEN
	  candq := candq || ',';
	END IF;	
	candq := candq || 'CNode('''||nnode.nview||''','''||nnode.rule||''','||REPLACE(TO_CHAR(nnode.entrop),',','.')||','||TO_CHAR(nnode.pop)||')';
		                   
      END LOOP; -- On att_values
      
      CLOSE att_values; 
      
      -- Actual candidate insertion

      egain := ROUND(egain,8); -- Precision adaptation
      -- Debug DBMS_OUTPUT.PUT_LINE('egain='||TO_CHAR(egain));
      candq := 'INSERT INTO BTCandidate VALUES (''' || catt || ''', '|| REPLACE(TO_CHAR(egain),',','.') ||', CNodeList(' || candq || '))';
      -- Debug DBMS_OUTPUT.PUT_LINE(candq);
      EXECUTE IMMEDIATE candq;

      
    END LOOP; -- On att_names
    
    CLOSE att_names; 
    
    -- Search for maximum information gain in candidate list
    
    maxg := 0;
    max_var := '';
    OPEN candidates FOR 'SELECT att_name, gain FROM BTCandidate';
    LOOP
      FETCH candidates INTO cvar, egain;
      EXIT WHEN candidates%NOTFOUND; -- Exit loop when last row is fetched    
      IF egain > maxg THEN
        maxg := egain;
        max_var := cvar;
      END IF;
    END LOOP;
    CLOSE candidates;
    -- Debug DBMS_OUTPUT.PUT_LINE('max_var='||max_var);
    
    -- Stack nodes for attribute with maximum infortmation gain superior to the minimum
    -- Drop the others
    
    OPEN candidates FOR 'SELECT * FROM BTCandidate';
    LOOP
      FETCH candidates INTO cvar, egain, elist;
      EXIT WHEN candidates%NOTFOUND; -- Exit loop when last row is fetched    
      IF (cvar <> max_var) OR (egain <= min_gain) THEN
        -- Destroy
        FOR i IN 1..elist.COUNT LOOP
          -- Debug DBMS_OUTPUT.PUT_LINE('destroying node '||elist(i).nview);
          EXECUTE IMMEDIATE 'DROP VIEW ' || elist(i).nview;
        END LOOP;
      ELSE
        FOR i IN 1..elist.COUNT LOOP
          -- Debug DBMS_OUTPUT.PUT_LINE('stacking node '||elist(i).nview);
          -- Push 
          stack.EXTEND;
          stack(stack.COUNT).num := nodenum;
          stack(stack.COUNT).nview := elist(i).nview;
          stack(stack.COUNT).entrop := elist(i).entrop;
          stack(stack.COUNT).pop := elist(i).pop;
          DBMS_OUTPUT.PUT_LINE('Creating node view ' || elist(i).nview);
          Result(nodenum, TO_CHAR(cnode.num), elist(i).rule, elist(i).nview); -- Output result
          nodenum := nodenum + 1; -- Next node
        END LOOP;          
      END IF; -- max
      elist.DELETE;
    END LOOP;
    CLOSE candidates; 	
    elist.TRIM(elist.COUNT); -- Empty
    
    -- Put current view in "delete" list
    
    delist.EXTEND;
    delist(delist.COUNT) := cnode.nview;  
    
/*  Debug  
    FOR i IN 1..stack.COUNT LOOP
      DBMS_OUTPUT.PUT_LINE('stack('||TO_CHAR(i)||').nview=' || stack(i).nview);
      DBMS_OUTPUT.PUT_LINE('stack('||TO_CHAR(i)||').entrop=' || TO_CHAR(stack(i).entrop));
    END LOOP;  
*/
  
  END LOOP; -- WHILE stack.COUNT > 0
  
  -- Cleaning...
  
  IF del THEN -- Suppress tree views
    FOR i IN 1..delist.COUNT LOOP
      -- DBMS_OUTPUT.PUT_LINE('Dropping node view ' || delist(i));
      EXECUTE IMMEDIATE 'DROP VIEW ' || delist(i);
    END LOOP;
  END IF;
  DBMS_OUTPUT.PUT_LINE('Dropping node views');
  delist.TRIM(delist.COUNT);
  classval.TRIM(classval.COUNT);
  
END;
/
